from datetime import datetime, timedelta
from typing import Any, Callable, List, Optional, Tuple
from pony import orm
import click

from laboris.model import Tag, Task
from laboris.ui import error, task_short
from laboris.util import date_parser as date, duration_parser as duration


def parse_goal(string: str) -> Tuple[timedelta, timedelta]:
    return duration(string.split("/", 1)[0]), duration(string.split("/", 1)[-1])


@click.command(name="create")
@click.option("-d", "--due", type=date, help="Set the due date of the task")
@click.option(
    "-P", "--priority", type=int, default=5, help="Set a priority for the task"
)
@click.option(
    "-p", "--parent", type=str, multiple=True, help="Specify any parent tasks"
)
@click.option(
    "-t", "--tag", type=str, multiple=True, help="Specify any tags for the task"
)
@click.option(
    "-g",
    "--goal",
    type=(duration, duration),
    help="Set a goal to work on the task",
)
@click.option(
    "-H",
    "--hidden/--no-hidden",
    default=False,
    help="Mark that task as hidden",
)
@click.argument("label", nargs=-1)
def create_command(
    label: Tuple[str, ...],
    due: Optional[datetime],
    priority: int,
    parent: Tuple[str, ...],
    tag: Tuple[str, ...],
    goal: Optional[Tuple[timedelta, timedelta]],
    hidden: bool,
):
    """Create a new task"""
    with orm.db_session:
        db_tags = list(Task.select())
        db_tasks = list(Task.select(lambda t: t.state == "OPEN"))

        arg_label = " ".join(
            x for x in label if x[0] not in ("+", "@") and ":" not in x
        )
        arg_tag = list(tag) + [x[1:] for x in label if x[0] == "@"]
        arg_parent = list(parent) + [x[1:] for x in label if x[0] == "+"]
        metadata = {x.split(":", 1)[0]: x.split(":", 1)[-1] for x in label if ":" in x}

        def parse_metadata(keys: List[str], func: Callable, default: Any) -> Any:
            for k in keys:
                if k in metadata:
                    return func(metadata[k])
            return default

        due = parse_metadata(["due", "dueDate"], date, due)
        priority = parse_metadata(["p", "pri", "priority"], int, priority)
        goal = parse_metadata(["goal", "target"], parse_goal, goal)

        tags = [
            next((y for y in db_tags if x.lower() == y.label.lower()), x)
            if type(x) == str
            else x
            for x in arg_tag
        ]
        tags = [Tag(label=x) if type(x) == str else x for x in tags]

        parents = [
            next(
                (
                    y
                    for y in db_tasks
                    if x.lower() == y.label.lower() or str(y.uuid).startswith(x)
                ),
                x,
            )
            if type(x) == str
            else x
            for x in arg_parent
        ]

        if any([type(x) == str for x in parents]):
            click.echo(
                error(
                    f"Not all parent tasks were found, missing {', '.join([repr(x) for x in parents if type(x) == str])}"
                )
            )
            return 1

        task = Task(
            label=arg_label,
            priority=priority,
            dueDate=due,
            tags=tags,
            goalDuration=goal[0] if goal else None,
            goalPeriod=goal[1] if goal else None,
            parents=parents,
            hidden=hidden,
        )

        click.echo(f"Created new task {task_short(task)}")



from typing import Any, Dict, List
import click
from click._compat import term_len

from laboris.model import Task


def error(*args, **kwargs):
    return click.style(*args, **kwargs, fg="red", bold=True)


def info(*args, **kwargs):
    return click.style(*args, **kwargs)


def term_width(string: str) -> int:
    return max(term_len(x) for x in string.split("\n"))


def task_short(tsk: Task):
    return (
        click.style(tsk.label, fg="blue")
        + " "
        + click.style("(" + tsk.shortUuid + ")", fg="black", bold=True)
    )


class Table(list):
    def __init__(
        self,
        header: List[str] = [],
        align: List[str] = [],
        header_style: Dict[str, Any] = {},
    ):
        self.header = header
        self.align = align
        self.header_style = header_style
        super().__init__()

    @staticmethod
    def pad(string: str, width: int, align: str = "<") -> str:
        if align == "<":
            return string + (" " * (width - term_width(string)))
        elif align == ">":
            return (" " * (width - term_width(string))) + string
        else:
            return (
                (" " * ((width - term_width(string)) // 2))
                + string
                + (" " * ((width - term_width(string)) // 2))
            )

    def raw(self) -> List[List[str]]:
        if self.header:
            return [self.header, *self]
        else:
            return self

    def get_widths(self) -> List[int]:
        widths: Dict[int, int] = {}
        for row in self:
            for idx, col in enumerate(row):
                widths[idx] = max(widths.get(idx, 0), term_width(col))

        if len(self.header) != 0:
            for idx, col in enumerate(self.header):
                widths[idx] = max(widths.get(idx, 0), term_width(col))

        return list(y for _, y in sorted(widths.items()))

    def format(self, csv: bool = False) -> str:
        widths = self.get_widths()

        if csv:
            return "\n".join(",".join(cell for cell in row) for row in self.raw())
        raw = self.raw()
        for i, row in enumerate(raw):
            raw[i] = [
                Table.pad(x, widths[j], self.align[j] if len(self.align) > j else "<")
                for j, x in enumerate(row)
            ]

        if len(self.header) != 0:
            raw[0] = [click.style(cell, **self.header_style) for cell in raw[0]]

        result = ""

        for row in raw:
            if any("\n" in x for x in row):
                lcount = max([x.count("\n") for x in row])
                for i in range(lcount + 1):
                    if i != 0:
                        result += "\n"
                    for j, cell in enumerate(row):
                        lines = cell.split("\n")
                        if j != 0:
                            result += "  "
                        if len(lines) <= i:
                            result += " " * widths[j]
                        else:
                            result += lines[i]
                pass
            else:
                result += "  ".join(cell for cell in row)
            result += "\n"

        return result.strip()

    def str(self) -> str:
        return self.format()



from typing import Optional
from laboris.util import date_parser as date, pick_one_task
from pony import orm
from datetime import datetime

import click

from laboris.ui import error, task_short

@click.command(name="start", context_settings={"ignore_unknown_options": True})
@click.argument("label", nargs=1)
@click.argument("date", type=date, required=False)
def start_command(label: str, date: Optional[datetime]):
    """Start tracking activity on a task"""
    with orm.db_session:
        db_task = pick_one_task(label)
        if db_task is None:
            return 1
        elif len(db_task.events) != 0 and len(db_task.events[-1]) == 1:
            click.echo(
                error(f"The task {task_short(db_task)}") + error(" is already active")
            )
            return 1
        db_task.events.append(
            [date.timestamp() if date else datetime.now().timestamp()]
        )
        db_task.updatedDate = datetime.now()
        click.echo(f"Started progress on {task_short(db_task)}")



from typing import Any, Callable, List, Tuple, Optional
from pony import orm
from datetime import datetime, timedelta
import click

from laboris.model import Tag, Task
from laboris.ui import error, task_short
from laboris.util import date_parser as date, duration_parser as duration, pick_one_task


def parse_goal(string: str) -> Tuple[timedelta, timedelta]:
    return duration(string.split("/", 1)[0]), duration(string.split("/", 1)[-1])


@click.command(name="update")
@click.option("-d", "--due", type=date, help="Set the due date of the task")
@click.option("-P", "--priority", type=int, help="Set a priority for the task")
@click.option(
    "-p", "--parent", type=str, multiple=True, help="Specify any parent tasks"
)
@click.option(
    "-t", "--tag", type=str, multiple=True, help="Specify any tags for the task"
)
@click.option(
    "-g",
    "--goal",
    type=(duration, duration),
    help="Set a goal to work on the task",
)
@click.option(
    "-H",
    "--hidden/--no-hidden",
    help="Mark that task as hidden",
)
@click.option(
    "-l", "--label", "new_label", type=str, multiple=True, help="New label for the task"
)
@click.argument("label", nargs=1)
@click.argument("args", nargs=-1)
def update_command(
    label: str,
    due: Optional[datetime],
    priority: Optional[int],
    parent: Tuple[str, ...],
    tag: Tuple[str, ...],
    goal: Optional[Tuple[timedelta, timedelta]],
    hidden: Optional[bool],
    new_label: Tuple[str, ...],
    args: Tuple[str, ...],
):
    """Update an exiting task"""
    with orm.db_session:
        task = pick_one_task(label)
        if task is None:
            return 1

        db_tags = list(Tag.select())
        db_tasks = list(Task.select(lambda t: t.state == "OPEN"))

        arg_tag = list(tag) + [x[1:] for x in args if x[0] == "@"]
        arg_parent = list(parent) + [x[1:] for x in args if x[0] == "+"]
        metadata = {x.split(":", 1)[0]: x.split(":", 1)[-1] for x in args if ":" in x}

        def parse_metadata(keys: List[str], func: Callable, default: Any) -> Any:
            for k in keys:
                if k in metadata:
                    return func(metadata[k])
            return default

        due = parse_metadata(["due", "dueDate"], date, due)
        priority = parse_metadata(["p", "pri", "priority"], int, priority)
        goal = parse_metadata(["goal", "target"], parse_goal, goal)

        if len(arg_tag) != 0:
            tags = [
                next((y for y in db_tags if x.lower() == y.label.lower()), x)
                if type(x) == str
                else x
                for x in arg_tag
            ]
            task.tags = [Tag(label=x) if type(x) == str else x for x in tags]

        if len(arg_parent) != 0:
            parents = [
                next((y for y in db_tasks if x.lower() == y.label.lower()), x)
                if type(x) == str
                else x
                for x in arg_parent
            ]

            if any([type(x) == str for x in parents]):
                click.echo(
                    error(
                        f"Not all parent tasks were found, missing {', '.join([repr(x) for x in parents if type(x) == str])}"
                    )
                )
                return 1

            task.parents = parents

        if due is not None:
            task.dueDate = due

        if priority is not None:
            task.priority = priority

        if goal is not None:
            task.goalDuration = goal[0]
            task.goalPeriod = goal[1]

        if hidden is not None:
            task.hidden = hidden

        if len(new_label) != 0:
            task.label = " ".join(new_label)

        task.updatedDate = datetime.now()
        click.echo(f"Updated task {task_short(task)}")



from datetime import datetime
from typing import Tuple
from laboris.util import taskQuery
from pony import orm
import humanize
import click

from laboris.ui import Table


@click.command(name="list")
@click.option("--csv/--no-csv", default=False, help="Format output as a CSV")
@click.option(
    "--done/--no-done", default=False, help="Include completed tasks in the results"
)
@click.option(
    "--hidden/--no-hidden", default=False, help="Include hidden tasks in the results"
)
@click.argument("query", nargs=-1)
def list_command(query: Tuple[str, ...], csv: bool, done: bool, hidden: bool):
    """List all current tasks in an easy to read table"""
    with orm.db_session:
        db_query = taskQuery(
            lambda t: (hidden or t.hidden == False) and (done or t.state == "OPEN"),
            *query,
        )
        db_tasks = sorted(list(db_query), key=lambda x: x.urgency, reverse=True)

        table = Table(
            header=["A", "UUID", "P", "Due", "Tags", "Projects", "Label", "Urg"],
            header_style={"underline": True, "bold": True},
            align=["<", "<", ">", ">", "<", "<", "<", ">"],
        )

        for t in db_tasks:
            table.append(
                [
                    click.style("●" if t.active else "", fg="green", bold=True),
                    click.style(t.shortUuid, fg="bright_black"),
                    click.style(t.priority, **t.priorityColor),
                    click.style(
                        f"{'-' if t.dueDate < datetime.now() else ''}{humanize.naturaldelta(t.dueDate - datetime.now())}"
                        if t.dueDate is not None
                        else "",
                        fg="magenta",
                    ),
                    click.style(
                        ("@" if len(t.tags) != 0 else "")
                        + " @".join([x.label for x in t.tags]),
                        fg="yellow",
                    ),
                    click.style(
                        ("+" if len(t.parents) != 0 else "")
                        + " +".join([x.label for x in t.parents]),
                        fg="blue",
                    ),
                    t.label,
                    click.style(f"{t.urgency:3.1f}", **t.urgencyColor),
                ]
            )

        click.echo(table.format(csv))



from os import path
from typing import Any
import pkg_resources
import click

from laboris.reports.detail import detail_command
from laboris.commands.done import done_command
from laboris.commands.start import start_command
from laboris.commands.stop import stop_command
from laboris.commands.create import create_command
from laboris.commands.delete import delete_command
from laboris.commands.update import update_command
from laboris.reports.gantt import gantt_command
from laboris.reports.graph import graph_command
from laboris.reports.list import list_command
from laboris.model import db

__version__ = pkg_resources.get_distribution("laboris").version


@click.group(name="laboris")
@click.option(
    "-v", "--verbose", count=True, help="Increase the verbosity of log messages."
)
@click.version_option(__version__)
def main(verbose: int):
    """A personal task manager and time tracker."""

    db.bind(
        provider="sqlite",
        filename=path.join(click.get_app_dir("laboris"), "database.sqlite"),
        create_db=True,
    )
    db.generate_mapping(create_tables=True)


# Create

main.add_command(create_command)

# Read

main.add_command(detail_command)
main.add_command(list_command)
main.add_command(graph_command)
main.add_command(gantt_command)

# Update

main.add_command(done_command)
main.add_command(start_command)
main.add_command(stop_command)
main.add_command(update_command)

# Delete

main.add_command(delete_command)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['laboris', 'laboris.commands', 'laboris.reports']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'halo>=0.0.31,<0.0.32',
 'humanize>=4.0.0,<5.0.0',
 'isodate>=0.6.1,<0.7.0',
 'pony>=0.7.16,<0.8.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'rich>=11.2.0,<12.0.0']

entry_points = \
{'console_scripts': ['laboris = laboris.main:main']}

setup_kwargs = {
    'name': 'laboris',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'Arden Rasmussen',
    'author_email': 'ardenisthebest@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
